from .aa import AATest
from .ab import ABTest
from .base import (
    Experiment,
    OnRoleExperiment,
)
from .base_complex import GroupExperiment, CycledExperiment
from .homogeneity import HomogeneityTest
from .matching import Matching

__all__ = [
    "CycledExperiment",
    "GroupExperiment",
    "AATest",
    "ABTest",
    "Matching",
    "HomogeneityTest",
]



from itertools import product
from typing import Optional, List, Dict, Sequence, Any

from tqdm.auto import tqdm

from hypex.dataset import ExperimentData, Dataset
from hypex.dataset import TempGroupingRole
from hypex.executor import Executor
from hypex.experiments.base import Experiment
from hypex.reporters import Reporter, DatasetReporter


class ExperimentWithReporter(Experiment):
    def __init__(
        self,
        executors: Sequence[Executor],
        reporter: Reporter,
        transformer: Optional[bool] = None,
        key: str = "",
    ):
        super().__init__(executors, transformer, key)
        self.reporter = reporter

    def one_iteration(self, data: ExperimentData, key: str = ""):
        t_data = ExperimentData(data.ds)
        self.key = key
        t_data = super().execute(t_data)
        return self.reporter.report(t_data)

    def _set_result(self, data: ExperimentData, result: List[Dataset]):
        result = result[0].append(result[1:], True)
        return self._set_value(data, result)


class CycledExperiment(ExperimentWithReporter):
    def __init__(
        self,
        executors: List[Executor],
        reporter: DatasetReporter,
        n_iterations: int,
        transformer: Optional[bool] = None,
        key: str = "",
    ):
        super().__init__(executors, reporter, transformer, key)
        self.n_iterations: int = n_iterations

    def generate_params_hash(self) -> str:
        return f"{self.reporter.__class__.__name__} x {self.n_iterations}"

    def execute(self, data: ExperimentData) -> ExperimentData:
        result: List[Dataset] = [
            self.one_iteration(data, str(i)) for i in tqdm(range(self.n_iterations))
        ]
        return self._set_result(data, result)


class GroupExperiment(ExperimentWithReporter):
    def generate_params_hash(self) -> str:
        return f"GroupExperiment: {self.reporter.__class__.__name__}"

    def execute(self, data: ExperimentData) -> ExperimentData:
        group_field = data.ds.search_columns(TempGroupingRole(), tmp_role=True)
        result: List[Dataset] = [
            self.one_iteration(group_data, str(group))
            for group, group_data in tqdm(data.ds.groupby(group_field))
        ]
        return self._set_result(data, result)


class ParamsExperiment(ExperimentWithReporter):
    def __init__(
        self,
        executors: Sequence[Executor],
        reporter: DatasetReporter,
        params: Dict[type, Dict[str, Sequence[Any]]],
        transformer: Optional[bool] = None,
        key: str = "",
    ):
        super().__init__(executors, reporter, transformer, key)
        self._params = params
        self._flat_params: List[Dict[type, Dict[str, Any]]] = []

    def generate_params_hash(self) -> str:
        return f"ParamsExperiment: {self.reporter.__class__.__name__}"

    def _update_flat_params(self):
        classes = list(self._params)
        param_combinations = [
            list(
                product(
                    *[
                        product([parameter], values)
                        for parameter, values in class_params.items()
                    ]
                )
            )
            for class_params in self._params.values()
        ]
        new_flat_params = [
            {
                classes[i]: dict(param_combination[i])
                for i in range(len(param_combination))
            }
            for param_combination in product(*param_combinations)
        ]
        self._flat_params = new_flat_params

    @property
    def flat_params(self) -> List[Dict[type, Dict[str, Any]]]:
        return self._flat_params

    @property
    def params(self) -> Dict[type, Dict[str, Sequence[Any]]]:
        return self._params

    @params.setter
    def params(self, params: Dict[type, Dict[str, Sequence[Any]]]):
        self._params = params
        self._update_flat_params()

    def execute(self, data: ExperimentData) -> ExperimentData:
        results = []
        self._update_flat_params()
        for flat_param in self._flat_params:
            t_data = ExperimentData(data.ds)
            for executor in self.executors:
                executor.set_params(flat_param)
                t_data = executor.execute(t_data)
            report = self.reporter.report(t_data)
            results.append(report)
        return self._set_result(data, results)



import warnings
from typing import List, Literal, Union

from hypex.analyzers.matching import MatchingAnalyzer
from hypex.comparators.distances import MahalanobisDistance
from hypex.dataset import TreatmentRole, TargetRole
from hypex.executor import Executor
from hypex.experiments.base import Experiment
from hypex.ml.faiss import FaissNearestNeighbors
from hypex.operators.operators import MatchingMetrics, Bias
from hypex.ui.base import ExperimentShell
from hypex.ui.matching import MatchingOutput


class Matching(ExperimentShell):

    @staticmethod
    def _make_experiment(
        distance: Literal["mahalanobis", "l2"] = "mahalanobis",
        two_sides: bool = True,
        metric: Literal["atc", "att", "ate", "auto"] = "auto",
        bias_estimation: bool = True,
        quality_tests: Union[
            Literal["smd", "psi", "ks-test", "repeats", "auto"],
            List[Literal["smd", "psi", "ks-test", "repeats", "auto"]],
        ] = "auto",
    ) -> Experiment:
        distance_mapping = {
            "mahalanobis": MahalanobisDistance(grouping_role=TreatmentRole())
        }
        executors: List[Executor] = [
            FaissNearestNeighbors(grouping_role=TreatmentRole(), two_sides=two_sides)
        ]
        if bias_estimation:
            executors += [
                Bias(grouping_role=TreatmentRole(), target_roles=[TargetRole()]),
            ]
        if metric in ["atc", "ate"] and not two_sides:
            raise ValueError(f"Can not estimate {metric} while two_sides is False")
        executors += [
            MatchingMetrics(
                grouping_role=TreatmentRole(),
                target_roles=[TargetRole()],
                metric=metric,
            ),
            MatchingAnalyzer(),
        ]
        if quality_tests != "auto":
            warnings.warn("Now quality tests aren't supported yet")
        return Experiment(
            executors=(
                executors
                if distance == "l2"
                else [distance_mapping[distance]] + executors
            )
        )

    def __init__(
        self,
        distance: Literal["mahalanobis", "l2"] = "mahalanobis",
        two_sides: bool = True,
        metric: Literal["atc", "att", "ate", "auto"] = "auto",
        bias_estimation: bool = True,
        quality_tests: Union[
            Literal["smd", "psi", "ks-test", "repeats", "auto"],
            List[Literal["smd", "psi", "ks-test", "repeats", "auto"]],
        ] = "auto",
    ):
        super().__init__(
            experiment=self._make_experiment(
                distance, two_sides, metric, bias_estimation, quality_tests
            ),
            output=MatchingOutput(),
        )



import sys

from hypex.analyzers import ABAnalyzer, OneAAStatAnalyzer
from hypex.comparators import GroupDifference, GroupSizes, ATE, TTest, KSTest, UTest
from hypex.dataset import (
    ExperimentData,
    Arg1Role,
    Arg2Role,
    InfoRole,
    TargetRole,
    FeatureRole,
    GroupingRole,
    PreTargetRole,
    StatisticRole,
    StratificationRole,
    TreatmentRole,
    TempTreatmentRole,
    TempGroupingRole,
    TempTargetRole,
)
from hypex.experiments import (
    Experiment,
    OnRoleExperiment,
    GroupExperiment,
    CycledExperiment,
)
from hypex.reporters import OneAADictReporter
from hypex.transformers import Shuffle
from hypex.utils import ExperimentDataEnum, SpaceEnum

all_classes = [
    ABAnalyzer,
    OneAAStatAnalyzer,
    GroupDifference,
    GroupSizes,
    ATE,
    TTest,
    KSTest,
    UTest,
    Arg1Role,
    Arg2Role,
    InfoRole,
    TargetRole,
    FeatureRole,
    GroupingRole,
    PreTargetRole,
    StatisticRole,
    StratificationRole,
    TreatmentRole,
    TempTreatmentRole,
    TempGroupingRole,
    TempTargetRole,
    OnRoleExperiment,
    GroupExperiment,
    CycledExperiment,
    OneAADictReporter,
    Shuffle,
    ExperimentDataEnum,
    SpaceEnum,
]

spaces = {
    "additional": SpaceEnum.additional,
    "auto": SpaceEnum.auto,
    "data": SpaceEnum.data,
}


class Factory:
    def __init__(self, hypothesis):
        self.hypothesis = hypothesis

    def make_experiment(self, experiment):
        executors = []
        for key, items in experiment.items():
            class_ = getattr(sys.modules[__name__], key)
            if "executors" in items or "inner_executors" in items:
                item = "executors" if "executors" in items else "inner_executors"
                items[f"{item}"] = self.make_experiment(experiment[key][f"{item}"][0])
            if "role" in items or "grouping_role" in items:
                item = "role" if "role" in items else "grouping_role"
                items[f"{item}"] = getattr(
                    sys.modules[__name__], items[item] + "Role"
                )()
            if "space" in items:
                items["space"] = spaces.get(items["space"])
            items = {i: None if j == "None" else j for i, j in items.items()}
            executors.append(class_(**items))
        return executors

    def execute(self):
        experiment_data = ExperimentData(self.hypothesis.dataset)
        experiment = Experiment(
            executors=self.make_experiment(self.hypothesis.experiment)
        )
        return experiment_data, experiment



from hypex.analyzers.aa import OneAAStatAnalyzer
from hypex.comparators import GroupDifference, GroupSizes
from hypex.comparators import TTest, KSTest, Chi2Test
from hypex.dataset import TargetRole, TreatmentRole
from hypex.experiments.base import Experiment, OnRoleExperiment
from hypex.ui.base import ExperimentShell
from hypex.ui.homo import HomoOutput
from hypex.utils import SpaceEnum

HOMOGENEITY_TEST = Experiment(
    executors=[
        OnRoleExperiment(
            executors=[
                GroupSizes(grouping_role=TreatmentRole(), compare_by="groups"),
                GroupDifference(grouping_role=TreatmentRole(), compare_by="groups"),
                TTest(grouping_role=TreatmentRole(), compare_by="groups"),
                KSTest(grouping_role=TreatmentRole(), compare_by="groups"),
                Chi2Test(grouping_role=TreatmentRole(), compare_by="groups"),
            ],
            role=TargetRole(),
        ),
        OneAAStatAnalyzer(),
    ]
)


class HomogeneityTest(ExperimentShell):
    def __init__(self):
        super().__init__(
            experiment=HOMOGENEITY_TEST,
            output=HomoOutput(),
        )



from typing import Dict, List, Any

from hypex.comparators import KSTest, TTest, Chi2Test
from hypex.dataset import Dataset, ExperimentData, StatisticRole
from hypex.executor import Executor
from hypex.experiments.base_complex import ParamsExperiment
from hypex.reporters.aa import OneAADictReporter
from hypex.splitters import AASplitter, AASplitterWithStratification
from hypex.utils import BackendsEnum, ExperimentDataEnum, ID_SPLIT_SYMBOL


class OneAAStatAnalyzer(Executor):
    def _set_value(self, data: ExperimentData, value, key=None) -> ExperimentData:
        return data.set_value(ExperimentDataEnum.analysis_tables, self.id, value)

    def execute(self, data: ExperimentData) -> ExperimentData:
        analysis_tests: List[type] = [TTest, KSTest, Chi2Test]
        executor_ids = data.get_ids(
            analysis_tests, searched_space=ExperimentDataEnum.analysis_tables
        )

        analysis_data: Dict[str, float] = {}
        for class_, spaces in executor_ids.items():
            analysis_ids = spaces.get("analysis_tables", [])
            if len(analysis_ids) > 0:
                if len(analysis_ids) > 1:
                    t_data = data.analysis_tables[analysis_ids[0]].append(
                        [data.analysis_tables[k] for k in analysis_ids[1:]]
                    )
                else:
                    t_data = data.analysis_tables[analysis_ids[0]]
                # t_data.data.index = analysis_ids
                for field in ["p-value", "pass"]:
                    analysis_data[f"mean {class_} {field}"] = t_data[field].mean()
        analysis_data["mean test score"] = (
            analysis_data["mean TTest p-value"]
            + 2 * analysis_data["mean KSTest p-value"]
        )
        if "mean Chi2Test p-value" in analysis_data:
            analysis_data["mean test score"] += (
                2 * analysis_data["mean Chi2Test p-value"]
            )
            analysis_data["mean test score"] /= 5
        else:
            analysis_data["mean test score"] /= 3

        analysis_dataset = Dataset.from_dict(
            [analysis_data],
            {field: StatisticRole() for field in analysis_data},
            BackendsEnum.pandas,
        )

        return self._set_value(data, analysis_dataset)


class AAScoreAnalyzer(Executor):
    AA_SPLITER_CLASS_MAPPING = {
        class_.__name__: class_ for class_ in [AASplitter, AASplitterWithStratification]
    }

    # TODO: rename alpha
    def __init__(self, alpha: float = 0.05, key: str = ""):
        super().__init__(key=key)
        self.alpha = alpha
        self.__feature_weights = {}

    def _set_value(
        self, data: ExperimentData, value: Any, key: Any = None
    ) -> ExperimentData:
        return data.set_value(
            ExperimentDataEnum.analysis_tables,
            executor_id=self.id,
            key=self.key,
            value=value,
        )

    def _analyze_aa_score(
        self, data: ExperimentData, score_table: Dataset
    ) -> ExperimentData:
        search_flag = f"{ID_SPLIT_SYMBOL}p-value{ID_SPLIT_SYMBOL}"
        self.__feature_weights = {
            column: 1 - abs(self.alpha - score_table.loc[:, column].mean())
            for column in score_table.columns
            if search_flag in column
        }
        aa_scores = {
            class_.replace(f"{ID_SPLIT_SYMBOL}p-value", ""): value
            for class_, value in self.__feature_weights.items()
        }
        aa_passed = {
            class_: value >= (1 - self.alpha * 1.2)
            for class_, value in aa_scores.items()
        }
        result = Dataset.from_dict({"score": aa_scores, "pass": aa_passed}, roles={})
        self.key = "aa score"
        return self._set_value(data, result)

    def build_splitter_from_id(self, splitter_id: str):
        splitter_class = self.AA_SPLITER_CLASS_MAPPING.get(
            splitter_id[: splitter_id.find(ID_SPLIT_SYMBOL)]
        )
        if splitter_class is None:
            raise ValueError(f"{splitter_id} is not a valid splitter id")
        return splitter_class.build_from_id(splitter_id)

    def _get_best_split(
        self, data: ExperimentData, score_table: Dataset
    ) -> Dict[str, Any]:
        aa_split_scores = score_table.apply(
            lambda x: (
                (
                    (
                        (
                            sum(
                                x[key] * value
                                for key, value in self.__feature_weights.items()
                            )
                            / len(self.__feature_weights)
                        )
                        * 2
                    )
                    / 3
                )
                + x["mean test score"] / 3
            ),
            axis=1,
            role={"aa split score": StatisticRole()},
        )
        best_index = aa_split_scores.idxmax()
        score_dict = score_table.loc[best_index, :].transpose().to_records()[0]
        best_score_stat = OneAADictReporter.convert_flat_dataset(score_dict)
        self.key = "best split statistics"
        result = self._set_value(data, best_score_stat)
        return {"index": best_index, "data": result}

    def _set_best_split(
        self, data: ExperimentData, score_table: Dataset, best_index: int
    ) -> ExperimentData:
        self.key = "best splitter"
        best_splitter_id = score_table.loc[best_index, "splitter_id"].get_values(0, 0)
        result = data.set_value(
            ExperimentDataEnum.variables, self.id, best_splitter_id, self.key
        )
        best_splitter = self.build_splitter_from_id(best_splitter_id)
        best_splitter.save_groups = False
        best_splitter.constant_key = False
        best_splitter.key = "best"
        result = best_splitter.execute(result)
        return result

    def _analyze_best_split(
        self, data: ExperimentData, score_table: Dataset
    ) -> ExperimentData:
        best_split = self._get_best_split(data, score_table)
        return self._set_best_split(
            best_split["data"], score_table, best_split["index"]
        )

    def execute(self, data: ExperimentData) -> ExperimentData:
        score_table_id = data.get_one_id(
            ParamsExperiment, ExperimentDataEnum.analysis_tables, "AATest"
        )
        score_table = data.analysis_tables[score_table_id]

        data = self._analyze_aa_score(data, score_table)
        data = self._analyze_best_split(data, score_table)
        return data



from hypex.reporters.matching import MatchingDatasetReporter

from .base import Output
from ..dataset import Dataset, ExperimentData, AdditionalMatchingRole


class MatchingOutput(Output):
    resume: Dataset
    full_data: Dataset

    def __init__(self):
        super().__init__(resume_reporter=MatchingDatasetReporter())

    def _extract_full_data(self, experiment_data: ExperimentData):
        indexes = experiment_data.additional_fields[
            experiment_data.additional_fields.search_columns(AdditionalMatchingRole())[0]
        ]
        indexes.index = experiment_data.ds.index
        filtered_field = indexes.drop(
            indexes[indexes[indexes.columns[0]] == -1], axis=0
        )
        matched_data = experiment_data.ds.loc[
            list(map(lambda x: x[0], filtered_field.get_values()))
        ].rename({i: i + "_matched" for i in experiment_data.ds.columns})
        matched_data.index = filtered_field.index
        self.indexes = indexes
        self.full_data = experiment_data.ds.append(
            matched_data.reindex(experiment_data.ds.index), axis=1
        )

    def extract(self, experiment_data: ExperimentData):
        super().extract(experiment_data)
        self._extract_full_data(experiment_data)



import json
from typing import Optional, Union, Dict, Any

from jsonschema import validate  # type: ignore

from hypex.dataset import Dataset, default_roles, InfoRole
from hypex.factory.base import Factory


class Hypothesis:
    def __init__(self, config: Union[str, Dict[str, Any]]):
        if isinstance(config, str):
            with open(config, "rb") as file:
                opened_config = json.load(file)
        else:
            opened_config = config
        with open("hypex\\hypotheses\\schemes\\scheme.json", "rb") as file:
            self.scheme = json.load(file)
        self.config = opened_config
        self.dataset = self.config.get("dataset")
        self.experiment = self.config.get("experiment")
        self.report = self.config.get("report")
        self.validate_config()
        self._parse_config()

    def validate_config(self):
        validate(self.config, self.scheme)
        if (
            "data" in self.dataset.keys()
            and "path" not in self.dataset.keys()
            and not self.dataset["data"]["data"]
        ):
            raise ValueError("Data ot path to data must be added")
        # if len(self.dataset["roles"]["role_names"]) != len(
        #     self.dataset["roles"]["columns"]
        # ):
        #     raise ValueError(
        #         f"Invalid number of columns and role_names. Columns and role_names must have equal length.\n "
        #         f"role_names contains {len(self.dataset['roles']['role_names'])} values and columns contains {len(self.dataset['roles']['columns'])}"
        #     )

    def _parse_config(self):
        self.dataset = self._parse_dataset()

    def _parse_dataset(self):
        data = (
            self.dataset["data"]
            if "data" in self.dataset.keys()
            else self.dataset["path"]
        )
        roles = {}
        for column in self.dataset["columns"]:
            role = default_roles.get(column["role"].lower(), InfoRole())
            role.data_type = column["dataType"] if column.get("dataType") else None
            roles.update({column["name"]: role})
        return Dataset(data=data, roles=roles, backend=self.dataset["backend"])

    def to_json(self, file: Optional[str] = None):
        # return json.dumps(self.dataset.to_json(), self.experiment.to_json(), self.report.to_json())
        if file:
            with open(file, "w") as f:
                json.dump(
                    {"dataset": self.dataset.to_dict(), "experiment": {}, "report": {}},
                    f,
                    indent=4,
                )
        return json.dumps(
            {"dataset": self.dataset.to_dict(), "experiment": {}, "report": {}}
        )

    def execute(self):
        experiment_data, self.experiment = Factory(self).execute()
        return experiment_data, self.experiment



from .constants import ID_SPLIT_SYMBOL, NAME_BORDER_SYMBOL, NUMBER_TYPES_LIST
from .enums import (
    SpaceEnum,
    BackendsEnum,
    ExperimentDataEnum,
    ABNTestMethodsEnum,
)
from .errors import (
    SpaceError,
    NoColumnsError,
    RoleColumnError,
    ConcatDataError,
    ConcatBackendError,
    NotFoundInExperimentDataError,
    FieldNotSuitableFieldError,
    DataTypeError,
    BackendTypeError,
    MergeOnError,
    AbstractMethodError,
    NoRequiredArgumentError,
)

from .typings import (
    FromDictTypes,
    TargetRoleTypes,
    DefaultRoleTypes,
    StratificationRoleTypes,
    CategoricalTypes,
    MultiFieldKeyTypes,
    DecoratedType,
    DocstringInheritDecorator,
    RoleNameType,
    ScalarType,
    SetParamsDictTypes,
)

__all__ = [
    "NAME_BORDER_SYMBOL",
    "ID_SPLIT_SYMBOL",
    "NUMBER_TYPES_LIST",
    "SpaceEnum",
    "BackendsEnum",
    "ExperimentDataEnum",
    "SpaceError",
    "NoColumnsError",
    "RoleColumnError",
    "ConcatDataError",
    "ConcatBackendError",
    "NotFoundInExperimentDataError",
    "FieldNotSuitableFieldError",
    "AbstractMethodError",
    "FromDictTypes",
    "TargetRoleTypes",
    "CategoricalTypes",
    "DefaultRoleTypes",
    "StratificationRoleTypes",
    "RoleNameType",
    "DecoratedType",
    "DocstringInheritDecorator",
    "MultiFieldKeyTypes",
    "DataTypeError",
    "BackendTypeError",
    "MergeOnError",
    "NoRequiredArgumentError",
    "ScalarType",
    "ABNTestMethodsEnum",
    "SetParamsDictTypes",
]



from typing import Optional, Literal, Union, List

from hypex.analyzers.ab import ABAnalyzer
from hypex.comparators import GroupDifference, GroupSizes, TTest, UTest, Chi2Test
from hypex.dataset import TargetRole, TreatmentRole
from hypex.experiments.base import Experiment, OnRoleExperiment
from hypex.ui.ab import ABOutput
from hypex.ui.base import ExperimentShell
from hypex.utils import ABNTestMethodsEnum


class ABTest(ExperimentShell):

    @staticmethod
    def _make_experiment(additional_tests, multitest_method):
        test_mapping = {
            "t-test": TTest(compare_by="groups", grouping_role=TreatmentRole()),
            "u-test": UTest(compare_by="groups", grouping_role=TreatmentRole()),
            "chi2-test": Chi2Test(compare_by="groups", grouping_role=TreatmentRole()),
        }
        on_role_executors = [GroupDifference(grouping_role=TreatmentRole())]
        additional_tests = ["t-test"] if additional_tests is None else additional_tests
        additional_tests = (
            additional_tests
            if isinstance(additional_tests, List)
            else [additional_tests]
        )
        for i in additional_tests:
            on_role_executors += [test_mapping[i]]
        return Experiment(
            executors=[
                GroupSizes(grouping_role=TreatmentRole()),
                OnRoleExperiment(
                    executors=on_role_executors,
                    role=TargetRole(),
                ),
                ABAnalyzer(
                    multitest_method=(
                        ABNTestMethodsEnum(multitest_method)
                        if multitest_method
                        else None
                    )
                ),
            ]
        )

    def __init__(
        self,
        additional_tests: Union[
            Literal["t-test", "u-test", "chi2-test"],
            List[Literal["t-test", "u-test", "chi2-test"]],
            None,
        ] = None,
        multitest_method: Optional[
            Literal[
                "bonferroni",
                "sidak",
                "holm-sidak",
                "holm",
                "simes-hochberg",
                "hommel",
                "fdr_bh",
                "fdr_by",
                "fdr_tsbh",
                "fdr_tsbhy",
                "quantile",
            ]
        ] = None,
    ):
        super().__init__(
            experiment=self._make_experiment(additional_tests, multitest_method),
            output=ABOutput(),
        )



from typing import Optional, Any, Dict

from hypex.comparators.distances import MahalanobisDistance
from hypex.dataset import Dataset, ABCRole, FeatureRole, ExperimentData, AdditionalMatchingRole
from hypex.executor import MLExecutor
from hypex.extensions.faiss import FaissExtension
from hypex.utils import SpaceEnum, ExperimentDataEnum
from hypex.utils.errors import PairsNotFoundError


class FaissNearestNeighbors(MLExecutor):
    def __init__(
        self,
        n_neighbors: int = 1,
        two_sides: bool = False,
        grouping_role: Optional[ABCRole] = None,
        space: SpaceEnum = SpaceEnum.auto,
        key: Any = "",
    ):
        self.n_neighbors = n_neighbors
        self.two_sides = two_sides
        super().__init__(
            grouping_role=grouping_role, target_role=FeatureRole(), key=key
        )

    @classmethod
    def _execute_inner_function(
        cls,
        grouping_data,
        target_field: Optional[str] = None,
        n_neighbors: Optional[int] = None,
        two_sides: Optional[bool] = None,
        **kwargs,
    ) -> Dict:
        data = cls._inner_function(
            data=grouping_data[0][1],
            test_data=grouping_data[1][1],
            n_neighbors=n_neighbors or 1,
            **kwargs,
        )
        if two_sides is None or two_sides == False:
            return {"test": data}
        return {
            "test": data,
            "control": cls._inner_function(
                data=grouping_data[1][1],
                test_data=grouping_data[0][1],
                n_neighbors=n_neighbors or 1,
                **kwargs,
            ),
        }

    @classmethod
    def _inner_function(
        cls,
        data: Dataset,
        test_data: Optional[Dataset] = None,
        target_data: Optional[Dataset] = None,
        n_neighbors: Optional[int] = None,
        **kwargs,
    ) -> Any:
        return FaissExtension(n_neighbors=n_neighbors or 1).calc(
            data=data, test_data=test_data
        )

    def fit(self, X: Dataset, Y: Optional[Dataset] = None) -> "MLExecutor":
        return FaissExtension(self.n_neighbors).fit(X=X, Y=Y)

    def predict(self, X: Dataset) -> Dataset:
        return FaissExtension().predict(X)

    def execute(self, data: ExperimentData) -> ExperimentData:
        group_field, features_fields = self._get_fields(data=data)
        if group_field[0] in data.groups:
            grouping_data = list(data.groups[group_field[0]].items())
        else:
            grouping_data = data.ds.groupby(group_field, fields_list=features_fields)
        distances_keys = data.get_ids(MahalanobisDistance, ExperimentDataEnum.groups)
        if len(distances_keys["MahalanobisDistance"]["groups"]) > 0:
            grouping_data = list(
                data.groups[distances_keys["MahalanobisDistance"]["groups"][0]].items()
            )
        compare_result = self.calc(
            data=data.ds,
            group_field=group_field,
            grouping_data=grouping_data,
            features_fields=features_fields,
            n_neighbors=self.n_neighbors,
            two_sides=self.two_sides,
        )
        ds = data.ds.groupby(group_field)
        matched_indexes = Dataset.create_empty()
        for i in range(len(compare_result.columns)):
            group = (
                grouping_data[1][1]
                if compare_result.columns[i] == "test"
                else grouping_data[0][1]
            )
            t_ds = ds[0][1] if compare_result.columns[i] == "test" else ds[1][1]
            t_index_field = (
                compare_result[compare_result.columns[i]]
                .loc[: len(group) - 1]
                .rename({compare_result.columns[i]: "indexes"})
            )
            if t_index_field.isna().sum() > 0:
                raise PairsNotFoundError
            matched_indexes = matched_indexes.append(
                Dataset.from_dict(
                    data={
                        "indexes": t_ds.iloc[
                            list(map(lambda x: int(x[0]), t_index_field.get_values()))
                        ].index
                    },
                    roles={"indexes": AdditionalMatchingRole()},
                    index=group.index,
                )
            ).sort()
        if len(matched_indexes) < len(data.ds) and not self.two_sides:
            matched_indexes = matched_indexes.reindex(data.ds.index, fill_value=-1)
        elif len(matched_indexes) < len(data.ds) and self.two_sides:
            raise PairsNotFoundError
        return self._set_value(data, matched_indexes, key="matched")



import datetime
from typing import Any, Union, Dict, List, TypeVar, Callable, Sequence

StratificationRoleTypes = Union[float, str, datetime.datetime]
DefaultRoleTypes = Union[float, bool, str, int]
TargetRoleTypes = Union[float, int, bool]
CategoricalTypes = Union[str]
ScalarType = Union[float, int, str, bool]


MultiFieldKeyTypes = Union[str, Sequence[str]]

FromDictTypes = Union[
    Dict[str, List[Any]], List[Dict[Any, Any]], Dict[str, Dict[Any, List]]
]
RoleNameType = str
DecoratedType = TypeVar("DecoratedType", bound=Union[Callable[..., Any], property])
DocstringInheritDecorator = Callable[[DecoratedType], DecoratedType]

SetParamsDictTypes = Union[Dict[str, Any], Dict[type, Dict[str, Any]]]



from hypex.analyzers.matching import MatchingAnalyzer
from hypex.dataset.dataset import ExperimentData
from hypex.reporters.abstract import DatasetReporter, DictReporter
from hypex.utils.enums import ExperimentDataEnum


class MatchingDictReporter(DictReporter):

    def _extract_from_analyser(self, data: ExperimentData):
        analyzer_id = data.get_one_id(
            MatchingAnalyzer, ExperimentDataEnum.analysis_tables
        )
        return self.extract_from_one_row_dataset(data.analysis_tables[analyzer_id])

    def report(self, experiment_data: ExperimentData):
        result = {}
        result.update(self._extract_from_analyser(experiment_data))
        return result


class MatchingDatasetReporter(DatasetReporter):
    def __init__(self) -> None:
        self.dict_reporter = MatchingDictReporter()



from typing import Dict, Optional, List, Literal, Union

from hypex.comparators.abstract import Comparator
from hypex.dataset import Dataset, ABCRole
from hypex.utils.constants import NUMBER_TYPES_LIST


class GroupDifference(Comparator):
    def __init__(
        self,
        compare_by: Literal[
            "groups", "columns", "columns_in_groups", "cross"
        ] = "groups",
        grouping_role: Optional[ABCRole] = None,
        target_roles: Union[ABCRole, List[ABCRole], None] = None,
    ):
        super().__init__(
            compare_by=compare_by,
            grouping_role=grouping_role,
            target_roles=target_roles,
        )

    @property
    def search_types(self) -> Optional[List[type]]:
        return NUMBER_TYPES_LIST

    @classmethod
    def _inner_function(
        cls,
        data: Dataset,
        test_data: Optional[Dataset] = None,
        **kwargs,
    ) -> Dict:
        test_data = cls._check_test_data(test_data)
        control_mean = data.mean()
        test_mean = test_data.mean()

        return {
            "control mean": control_mean,
            "test mean": test_mean,
            "difference": test_mean - control_mean,
            "difference %": (test_mean / control_mean - 1) * 100,
        }


class GroupSizes(Comparator):
    def __init__(
        self,
        compare_by: Literal[
            "groups", "columns", "columns_in_groups", "cross"
        ] = "groups",
        grouping_role: Optional[ABCRole] = None,
    ):
        super().__init__(
            compare_by=compare_by,
            grouping_role=grouping_role,
            target_roles=grouping_role,
        )

    @classmethod
    def _inner_function(
        cls, data: Dataset, test_data: Optional[Dataset] = None, **kwargs
    ) -> Dict:
        size_a = len(data)
        size_b = len(test_data) if isinstance(test_data, Dataset) else 0

        return {
            "control size": size_a,
            "test size": size_b,
            "control size %": (size_a / (size_a + size_b)) * 100,
            "test size %": (size_b / (size_a + size_b)) * 100,
        }



import warnings
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Union, Tuple, Sequence, Literal

from hypex.dataset import (
    ABCRole,
    Dataset,
    ExperimentData,
    StatisticRole,
    TempTargetRole,
    InfoRole,
    DatasetAdapter,
    GroupingRole,
    TargetRole,
    PreTargetRole,
)
from hypex.executor import Calculator
from hypex.utils import (
    BackendsEnum,
    ExperimentDataEnum,
    FromDictTypes,
    NAME_BORDER_SYMBOL,
)
from hypex.utils.adapter import Adapter
from hypex.utils.errors import (
    AbstractMethodError,
    FieldNotSuitableFieldError,
    NoRequiredArgumentError,
    NoColumnsError,
)


class Comparator(Calculator, ABC):
    def __init__(
        self,
        compare_by: Literal["groups", "columns", "columns_in_groups", "cross"],
        grouping_role: Optional[ABCRole] = None,
        target_roles: Union[ABCRole, List[ABCRole], None] = None,
        baseline_role: Optional[ABCRole] = None,
        key: Any = "",
    ):
        super().__init__(key=key)
        self.grouping_role = grouping_role or GroupingRole()
        self.compare_by = compare_by
        self.target_roles = target_roles or TargetRole()
        self.baseline_role = baseline_role or PreTargetRole()

    @property
    def search_types(self) -> Optional[List[type]]:
        return None

    def _local_extract_dataset(
        self, compare_result: Dict[Any, Any], roles: Dict[Any, ABCRole]
    ) -> Dataset:
        return self._extract_dataset(compare_result, roles)

    @classmethod
    @abstractmethod
    def _inner_function(
        cls, data: Dataset, test_data: Optional[Dataset] = None, **kwargs
    ) -> Any:
        raise AbstractMethodError

    def _get_fields(self, data: ExperimentData) -> Dict[str, Union[str, List[str]]]:
        tmp_role = True if data.ds.tmp_roles else False
        group_field = self._field_searching(
            data=data,
            roles=self.grouping_role,
        )
        target_fields = self._field_searching(
            data=data,
            roles=TempTargetRole() if tmp_role else self.target_roles,
            tmp_role=tmp_role,
            search_types=self.search_types,
        )
        baseline_field = self._field_searching(
            data=data,
            roles=self.baseline_role,
            tmp_role=tmp_role,
        )
        return {
            "group_field": group_field,
            "target_fields": target_fields,
            "baseline_field": baseline_field,
        }

    @classmethod
    def _execute_inner_function(
        cls,
        baseline_data: List[Tuple[str, Dataset]],
        compared_data: List[Tuple[str, Dataset]],
        **kwargs,
    ) -> Dict:
        result = {}
        for i in range(len(compared_data)):
            result[compared_data[i][0]] = DatasetAdapter.to_dataset(
                cls._inner_function(
                    baseline_data[0 if len(baseline_data) == 1 else i][1],
                    compared_data[i][1],
                    **kwargs,
                ),
                InfoRole(),
            )
        return result

    @staticmethod
    def _check_test_data(test_data: Optional[Dataset] = None) -> Dataset:
        if test_data is None:
            raise ValueError("test_data is needed for evaluation")
        return test_data

    def _set_value(
        self, data: ExperimentData, value: Optional[Dataset] = None, key: Any = None
    ) -> ExperimentData:
        data.set_value(
            ExperimentDataEnum.analysis_tables,
            self.id,
            value,
        )
        return data

    @staticmethod
    def _extract_dataset(
        compare_result: FromDictTypes, roles: Dict[Any, ABCRole]
    ) -> Dataset:
        if isinstance(list(compare_result.values())[0], Dataset):
            cr_list_v: List[Dataset] = list(compare_result.values())
            result = cr_list_v[0]
            if len(cr_list_v) > 1:
                result = result.append(cr_list_v[1:])
            result.index = list(compare_result.keys())
            return result
        return Dataset.from_dict(compare_result, roles, BackendsEnum.pandas)

    # TODO выделить в отдельную функцию с кваргами (нужно для альфы)

    @staticmethod
    def _grouping_data_split(
        grouping_data: Union[Tuple[List[Tuple[str, Dataset]]], Dict[str, Dataset]],
        compare_by: Literal["groups", "columns", "columns_in_groups", "cross"],
        target_fields: List[str],
        baseline_field: Optional[str],
    ):
        if isinstance(grouping_data, Tuple):
            baseline_data = Adapter.to_list(grouping_data[0])
            compared_data = Adapter.to_list(grouping_data[1])
        elif isinstance(grouping_data, Dict):
            compared_data = [(item[0], item[1]) for item in grouping_data.items()]
            baseline_data = [compared_data.pop()]
        else:
            raise TypeError(
                f"Grouping data must be tuple or list or tuple of lists, but got {type(grouping_data)}"
            )

        baseline_data = [
            (
                bucket[0],
                bucket[1][target_fields if compare_by == "groups" else baseline_field],
            )
            for bucket in baseline_data
        ]
        compared_data = [
            (bucket[0], bucket[1][target_fields]) for bucket in compared_data
        ]

        return baseline_data, compared_data

    @staticmethod
    def _split_ds_into_columns(
        data: List[Tuple[str, Dataset]]
    ) -> List[Tuple[str, Dataset]]:
        result = [
            (f"{bucket[0]}{NAME_BORDER_SYMBOL}{column}", bucket[1][column])
            for bucket in data
            for column in bucket[1].columns
        ]
        return result

    @classmethod
    def _split_for_groups_mode(
        cls, data: Dataset, group_field: str, target_fields: List[str]
    ):
        if isinstance(target_fields, List) and len(target_fields) > 1:
            warnings.warn(
                f"Only one Target field can be passed when the comparison is done by groups. {len(target_fields)} passed. {target_fields[0]} will be used.",
            )
            target_fields = target_fields[0]
        if not isinstance(group_field, str):
            raise TypeError(
                f"group_field must be one string, {type(group_field)} passed."
            )

        data_buckets = sorted(data.groupby(by=group_field, fields_list=target_fields))
        baseline_data = cls._split_ds_into_columns([data_buckets.pop(0)])
        compared_data = cls._split_ds_into_columns(data=data_buckets)
        return baseline_data, compared_data

    @classmethod
    def _split_for_columns_mode(
        cls, data: Dataset, baseline_field: Optional[str], target_fields: List[str]
    ):
        if baseline_field is None:
            raise NoRequiredArgumentError("baseline_field")

        baseline_data = [(f"{baseline_field}", data[baseline_field])]
        compared_data = [(f"{column}", data[column]) for column in target_fields]
        return baseline_data, compared_data

    @classmethod
    def _split_for_columns_in_groups_mode(
        cls,
        data: Dataset,
        group_field: Optional[str],
        baseline_field: Optional[str],
        target_fields: List[str],
    ):
        if baseline_field is None:
            raise NoRequiredArgumentError("baseline_field")
        if group_field is None:
            raise NoRequiredArgumentError("group_field")
        if not isinstance(group_field, str):
            raise TypeError(
                f"group_field must be one string, {type(group_field)} passed."
            )
        if len(target_fields) > 1:
            warnings.warn(
                f"Too many fields passed as Target fields {target_fields}, only {target_fields[0]} will be used"
            )

        baseline_data = cls._split_ds_into_columns(
            data.groupby(by=group_field, fields_list=baseline_field)
        )
        compared_data = data.groupby(by=group_field, fields_list=target_fields[0])
        compared_data = cls._split_ds_into_columns(data=compared_data)
        return baseline_data, compared_data

    @classmethod
    def _split_for_cross_mode(
        cls,
        data: Dataset,
        group_field: Optional[str],
        baseline_field: Optional[str],
        target_fields: List[str],
    ):
        if baseline_field is None:
            raise NoRequiredArgumentError("baseline_field")
        if group_field is None:
            raise NoRequiredArgumentError("group_field")
        if not isinstance(group_field, str):
            raise TypeError(
                f"group_field must be one string, {type(group_field)} passed."
            )
        data_buckets = sorted(data.groupby(by=group_field, fields_list=baseline_field))
        baseline_data = cls._split_ds_into_columns([data_buckets.pop(0)])
        compared_data = sorted(data.groupby(by=group_field, fields_list=target_fields))
        compared_data.pop(0)
        compared_data = cls._split_ds_into_columns(data=compared_data)
        return baseline_data, compared_data

    @classmethod
    def _split_data_to_buckets(
        cls,
        data: Dataset,
        compare_by: Literal["groups", "columns", "columns_in_groups", "cross"],
        target_fields: Union[str, List[str]],
        baseline_field: Optional[str] = None,
        group_field: Optional[str] = None,
    ) -> Tuple:
        """
        Splits the given dataset into buckets into baseline and compared data, based on the specified comparison mode.

        Args:
            data (Dataset): The dataset to be split.
            group_field (Union[Sequence[str], str]): The field(s) to group the data by.
            target_fields (Union[str, List[str]]): The field(s) to target for comparison.
            compare_by (Literal['groups', 'columns', 'columns_in_groups', 'cross'], optional): The method to compare the data. Defaults to 'groups'.
            baseline_field (Optional[str], optional): The column to use as the baseline for comparison. Required if `compare_by` is 'columns' or 'columns_in_groups'. Defaults to None.

        Returns:
            Tuple: A tuple containing the baseline data and the compared data.

        Raises:
            NoRequiredArgumentError: If `baseline_field` is None and `compare_by` is 'columns' or 'columns_in_groups' or 'cross'.
            ValueError: If `compare_by` is not one of the allowed values.
        """

        target_fields = Adapter.to_list(target_fields)
        if compare_by == "groups":
            baseline_data, compared_data = cls._split_for_groups_mode(
                data, group_field, target_fields
            )
        elif compare_by == "columns":
            baseline_data, compared_data = cls._split_for_columns_mode(
                data, baseline_field, target_fields
            )
        elif compare_by == "columns_in_groups":
            baseline_data, compared_data = cls._split_for_columns_in_groups_mode(
                data, group_field, baseline_field, target_fields
            )
        elif compare_by == "cross":
            baseline_data, compared_data = cls._split_for_cross_mode(
                data, group_field, baseline_field, target_fields
            )
        else:
            raise ValueError(
                f"Wrong compare_by argument passed {compare_by}. It can be only one of the following modes: 'groups', 'columns', 'columns_in_groups', 'cross'."
            )
        return baseline_data, compared_data

    @classmethod
    def calc(
        cls,
        data: Dataset,
        compare_by: Literal["groups", "columns", "columns_in_groups", "cross"],
        target_fields: Union[str, List[str]],
        baseline_field: Optional[str] = None,
        group_field: Optional[str] = None,
        grouping_data: Optional[Tuple[List[Tuple[str, Dataset]]]] = None,
        **kwargs,
    ) -> Dict:
        target_fields = Adapter.to_list(target_fields)

        if grouping_data is None:
            grouping_data = cls._split_data_to_buckets(
                data=data,
                compare_by=compare_by,
                target_fields=target_fields,
                baseline_field=baseline_field,
                group_field=group_field,
            )

        if len(grouping_data[0]) < 1 or len(grouping_data[1]) < 1:
            raise FieldNotSuitableFieldError(group_field, "Grouping")

        baseline_data, compared_data = grouping_data
        return cls._execute_inner_function(
            baseline_data=baseline_data,
            compared_data=compared_data,
            **kwargs,
        )

    def execute(self, data: ExperimentData) -> ExperimentData:
        fields = self._get_fields(data)

        group_field = fields["group_field"]
        target_fields = fields["target_fields"]
        baseline_field = fields["baseline_field"]

        self.key = str(
            target_fields[0] if len(target_fields) == 1 else (target_fields or "")
        )
        if not target_fields:
            if (
                data.ds.tmp_roles
            ):  # если колонка не подходит для теста, то тагет будет пустой, но если есть темп роли, то это нормальное поведение
                return data
            else:
                raise NoColumnsError(TargetRole().role_name)

        if not group_field and self.compare_by != "columns":
            raise FieldNotSuitableFieldError(group_field, "Grouping")

        if (
            group_field[0] in data.groups
        ):  # TODO: propper split between groups and columns
            grouping_data = self._grouping_data_split(
                grouping_data=data.groups[group_field[0]],
                compare_by=self.compare_by,
                target_fields=target_fields,
                baseline_field=baseline_field,
            )
        else:
            grouping_data = None
            data.groups[group_field[0]] = {
                f"{group}": ds for group, ds in data.ds.groupby(group_field[0])
            }

        compare_result = self.calc(
            data=data.ds,
            compare_by=self.compare_by,
            target_fields=target_fields,
            baseline_field=Adapter.list_to_single(baseline_field),
            group_field=Adapter.list_to_single(group_field),
            grouping_data=grouping_data,
        )
        result_dataset = self._local_extract_dataset(
            compare_result, {key: StatisticRole() for key in compare_result}
        )
        return self._set_value(data, result_dataset)


class StatHypothesisTesting(Comparator, ABC):
    def __init__(
        self,
        compare_by: Literal["groups", "columns", "columns_in_groups", "cross"],
        grouping_role: Union[ABCRole, None] = None,
        target_role: Union[ABCRole, None] = None,
        baseline_role: Union[ABCRole, None] = None,
        reliability: float = 0.05,
        key: Any = "",
    ):
        super().__init__(
            compare_by=compare_by,
            grouping_role=grouping_role,
            target_roles=target_role,
            baseline_role=baseline_role,
            key=key,
        )
        self.reliability = reliability



from copy import deepcopy
from typing import Optional, Any, List, Literal, Union, Dict, Tuple

import numpy as np

from hypex.dataset import (
    Dataset,
    ABCRole,
    ExperimentData,
    TargetRole,
    InfoRole,
    FeatureRole,
    AdditionalMatchingRole,
    AdditionalTargetRole,
)
from hypex.operators.abstract import GroupOperator
from hypex.utils.enums import ExperimentDataEnum
from hypex.utils.errors import NoneArgumentError


class SMD(GroupOperator):
    def execute(self, data: ExperimentData) -> ExperimentData:
        pass

    @classmethod
    def _inner_function(
        cls, data: Dataset, test_data: Optional[Dataset] = None, **kwargs
    ) -> Any:
        test_data = cls._check_test_data(test_data=test_data)
        return (data.mean() + test_data.mean()) / data.std()


class MatchingMetrics(GroupOperator):
    def __init__(
        self,
        grouping_role: Optional[ABCRole] = None,
        target_roles: Union[ABCRole, List[ABCRole], None] = None,
        metric: Optional[Literal["auto", "atc", "att", "ate"]] = None,
        key: Any = "",
    ):
        self.metric = metric or "auto"
        target_roles = target_roles or TargetRole()
        super().__init__(
            grouping_role=grouping_role,
            target_roles=(
                target_roles if isinstance(target_roles, List) else [target_roles]
            ),
            key=key,
        )

    @classmethod
    def _inner_function(
        cls,
        data: Dataset,
        test_data: Optional[Dataset] = None,
        target_fields: Optional[List[str]] = None,
        **kwargs,
    ) -> Any:
        if target_fields is None or test_data is None:
            raise NoneArgumentError(
                ["target_fields", "test_data"], "att, atc, ate estimation"
            )
        metric = kwargs.get("metric", "ate")
        itt = test_data[target_fields[0]] - test_data[target_fields[1]]
        itc = data[target_fields[1]] - data[target_fields[0]]
        bias = kwargs.get("bias", {})
        if len(bias) > 0:
            if metric in ["atc", "ate"]:
                itc -= Dataset.from_dict({"test": bias["control"]}, roles={})
            itt += Dataset.from_dict({"control": bias["test"]}, roles={})
        itt = itt.mean()
        itc = itc.mean()
        if metric == "atc":
            return {"ATC": itc}
        if metric == "att":
            return {"ATT": itt}
        len_test, len_control = len(data), len(test_data)
        return {
            "ATT": itt,
            "ATC": itc,
            "ATE": (itt * len_test + itc * len_control) / (len_test + len_control),
        }

    @staticmethod
    def _define_metric(
        grouping_data: List[Tuple[Union[int, str], Dataset]], target_fields
    ):
        if grouping_data[0][1][target_fields[1]].isna().sum() == len(
            grouping_data[0][1]
        ):
            return "att"
        else:
            return (
                "atc"
                if grouping_data[0][1][target_fields[1]].isna().sum()
                == len(grouping_data[1][1])
                else "ate"
            )

    @classmethod
    def _execute_inner_function(
        cls, grouping_data, target_fields: Optional[List[str]] = None, **kwargs
    ) -> Dict:
        metric = kwargs.get("metric", "auto")
        if target_fields is None or len(target_fields) != 2:
            raise ValueError(
                "This operator works with 2 targets, but got {}".format(
                    len(target_fields) if target_fields else None
                )
            )
        if metric == "auto":
            metric = cls._define_metric(grouping_data, target_fields)
        return cls._inner_function(
            data=grouping_data[0][1],
            test_data=grouping_data[1][1],
            target_fields=target_fields,
            metric=metric,
            bias=kwargs.get("bias_estimation", None),
        )

    def _prepare_new_target(self, data: ExperimentData, t_data: Dataset) -> Dataset:
        indexes = self._field_searching(data, AdditionalMatchingRole())
        if len(indexes) == 0:
            raise ValueError(f"No indexes were found")
        new_target = data.ds.search_columns(TargetRole())[0]
        indexes = data.additional_fields[indexes[0]]
        indexes.index = t_data.index
        filtered_field = indexes.drop(
            indexes[indexes[indexes.columns[0]] == -1], axis=0
        )
        matched_data = data.ds.loc[
            list(map(lambda x: x[0], filtered_field.get_values()))
        ][new_target].rename(
            {new_target: new_target + "_matched" for _ in data.ds.columns}
        )
        matched_data.index = filtered_field.index
        return matched_data

    def execute(self, data: ExperimentData) -> ExperimentData:
        group_field, target_fields = self._get_fields(data=data)
        bias = (
            data.variables[data.get_one_id(Bias, ExperimentDataEnum.variables)]
            if len(
                data.get_ids(Bias, ExperimentDataEnum.variables)["Bias"]["variables"]
            )
            > 0
            else None
        )
        t_data = deepcopy(data.ds)
        if len(target_fields) != 2:
            matched_data = self._prepare_new_target(data, t_data)
            target_fields += [matched_data.search_columns(TargetRole())[0]]
            data.set_value(
                ExperimentDataEnum.additional_fields,
                self.id,
                matched_data,
                role=AdditionalTargetRole(),
            )
            t_data = t_data.add_column(
                matched_data.reindex(t_data.index),
                role={target_fields[1]: TargetRole()},
            )
        self.key = str(
            target_fields[0] if len(target_fields) == 1 else (target_fields or "")
        )
        if (
            not target_fields and data.ds.tmp_roles
        ):  # если колонка не подходит для теста, то тагет будет пустой, но если есть темп роли, то это нормальное поведение
            return data

        compare_result = self.calc(
            data=t_data,
            group_field=group_field,
            target_fields=target_fields,
            metric=self.metric,
            bias_estimation=bias,
        )
        return self._set_value(data, compare_result)


class Bias(GroupOperator):
    def __init__(
        self,
        grouping_role: Optional[ABCRole] = None,
        target_roles: Optional[List[ABCRole]] = None,
        key: Any = "",
    ):
        super().__init__(
            grouping_role=grouping_role, target_roles=target_roles, key=key
        )

    @staticmethod
    def calc_coefficients(X: Dataset, Y: Dataset) -> List[float]:
        X_l = Dataset.create_empty(roles={"temp": InfoRole()}, index=X.index).fillna(1)
        return np.linalg.lstsq(
            X_l.append(X, axis=1).data.values, Y.data.values, rcond=-1
        )[0][1:]

    @staticmethod
    def calc_bias(
        X: Dataset, X_matched: Dataset, coefficients: List[float]
    ) -> List[float]:
        return [
            (j - i).dot(coefficients)[0]
            for i, j in zip(X.data.values, X_matched.data.values)
        ]

    @classmethod
    def _inner_function(
        cls,
        data: Dataset,
        test_data: Optional[Dataset] = None,
        target_fields: Optional[List[str]] = None,
        features_fields: Optional[List[str]] = None,
        **kwargs,
    ) -> Dict:
        if target_fields is None or features_fields is None or test_data is None:
            raise NoneArgumentError(
                ["target_fields", "features_fields", "test_data"], "bias_estimation"
            )
        test_result = cls.calc_bias(
            test_data[features_fields[: len(features_fields) // 2]],
            test_data[features_fields[len(features_fields) // 2 :]],
            cls.calc_coefficients(
                test_data[features_fields[len(features_fields) // 2 :]],
                test_data[target_fields[1]],
            ),
        )
        if data[target_fields[1]].isna().sum() > 0:
            return {"test": test_result}
        return {
            "test": test_result,
            "control": cls.calc_bias(
                data[features_fields[: len(features_fields) // 2]],
                data[features_fields[len(features_fields) // 2 :]],
                cls.calc_coefficients(
                    data[features_fields[len(features_fields) // 2 :]],
                    data[target_fields[1]],
                ),
            ),
        }

    @classmethod
    def _execute_inner_function(
        cls,
        grouping_data,
        target_fields: Optional[List[str]] = None,
        features_fields: Optional[List[str]] = None,
        **kwargs,
    ) -> Dict:
        return cls._inner_function(
            grouping_data[0][1],
            test_data=grouping_data[1][1],
            target_fields=target_fields,
            features_fields=features_fields,
            **kwargs,
        )

    def _prepare_data(self, data: ExperimentData, t_data: Dataset) -> Dataset:
        indexes = data.additional_fields[
            self._field_searching(data, AdditionalMatchingRole())[0]
        ]
        indexes.index = t_data.index
        filtered_field = indexes.drop(
            indexes[indexes[indexes.columns[0]] == -1], axis=0
        )
        matched_data = data.ds.loc[
            list(map(lambda x: x[0], filtered_field.get_values()))
        ].rename({i: i + "_matched" for i in data.ds.columns})
        matched_data.index = filtered_field.index
        return matched_data

    def execute(self, data: ExperimentData) -> ExperimentData:
        group_field, target_fields = self._get_fields(data)
        t_data = deepcopy(data.ds)
        if len(target_fields) < 2:
            matched_data = self._prepare_data(data, t_data)
            target_fields += [matched_data.search_columns(TargetRole())[0]]
            t_data = t_data.append(matched_data.reindex(t_data.index), axis=1)
        self.key = str(
            target_fields[0] if len(target_fields) == 1 else (target_fields or "")
        )
        if (
            not target_fields and data.ds.tmp_roles
        ):  # если колонка не подходит для теста, то тагет будет пустой, но если есть темп роли, то это нормальное поведение
            return data

        compare_result = self.calc(
            data=t_data,
            group_field=group_field,
            target_fields=target_fields,
            features_fields=t_data.search_columns(
                FeatureRole(), search_types=[int, float]
            ),
        )
        return self._set_value(data, compare_result)
